/*
 * Copyright 2020 Confluent Inc.
 *
 * Licensed under the Confluent Community License (the "License"); you may not use
 * this file except in compliance with the License.  You may obtain a copy of the
 * License at
 *
 * http://www.confluent.io/confluent-community-license
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 */

package io.confluent.ksql.test.model;

import io.confluent.ksql.model.WindowType;
import java.util.Optional;
import org.junit.Test;

public class KeyFormatNodeTest {

  static final KeyFormatNode INSTANCE = new KeyFormatNode(
      Optional.of("json"),
      Optional.of(WindowType.SESSION),
      Optional.of(42L)
  );

  @Test
  public void shouldRoundTrip() {
    ModelTester.assertRoundTrip(INSTANCE);
  }
}
